﻿using System;
using System.Configuration;
using System.IO;
using System.IO.Compression;
using System.Net;
using System.Text.RegularExpressions;
using System.Web;

namespace GerkePreussner.Blog.ReverseProxy
{
    /// <summary>
    /// Implements a reverse proxy HTTP handler for IIS.
    /// </summary>

    public class ReverseProxy : IHttpAsyncHandler
    {
        #region Constructors

        /// <summary>
        /// Creates a new instance of the HttpReverseProxy class and initializes the event log.
        /// </summary>

        public ReverseProxy ()
        {
            Log.Information("Reverse proxy created.");
        }

        #endregion


        #region IHttpAsyncHandler

        /// <summary>
        /// Initiates an asynchronous call to the HTTP handler.
        /// </summary>
        /// <param name="context">HTTP context object.</param>
        /// <param name="cb">Asynchronous callback object.</param>
        /// <param name="extraData">Extra data needed to process the request.</param>
        /// <returns>Asynchronous result object that contains the status of the process.</returns>

        public IAsyncResult BeginProcessRequest (HttpContext context, AsyncCallback cb, object extraData)
        {
            AsyncState state = new AsyncState();

            state.LocalContext = context;

            // construct remote URL
            state.LocalWebsite = context.Request.Url.Scheme + "://" + context.Request.Url.Authority;
            
            if (context.Request.ApplicationPath != "/")
            {              
                state.LocalWebsite += context.Request.ApplicationPath;
            }

            state.RemoteWebsite = ConfigurationManager.AppSettings.Get("RemoteWebsite");
            state.RemoteUrl = Regex.Replace(context.Request.Url.AbsoluteUri, state.LocalWebsite, state.RemoteWebsite, RegexOptions.IgnoreCase);

            Log.Information("Begin Request ---\nLocal Request: " + state.LocalContext.Request.Url + "\nLocal Website: " + state.LocalWebsite + "\nRemote Website: " + state.RemoteWebsite + "\nRemote Request: " + state.RemoteUrl);

            // prepare remote request
            state.RemoteRequest = WebRequest.Create(state.RemoteUrl) as HttpWebRequest;
            state.RemoteRequest.Accept = "*/*";
            state.RemoteRequest.AllowAutoRedirect = false;
            state.RemoteRequest.ContentType = context.Request.ContentType;
            state.RemoteRequest.Method = context.Request.HttpMethod;
            state.RemoteRequest.Proxy = null;
            state.RemoteRequest.UserAgent = context.Request.UserAgent;

            if (ConfigurationManager.AppSettings.Get("AllowRemoteCompression") == "true")
            {
                state.RemoteRequest.AutomaticDecompression = DecompressionMethods.GZip | DecompressionMethods.Deflate;
            }

            // forward headers
            foreach (string h in context.Request.Headers)
            {
                if (!WebHeaderCollection.IsRestricted(h) && (h != "Accept-Encoding") && (h != "Remote-User"))
                {
                    state.RemoteRequest.Headers.Add(h, context.Request.Headers.Get(h));
                }
            }

            state.RemoteRequest.Headers.Add("Remote-User", context.User.Identity.Name);

            // forward content
            if ((context.Request.HttpMethod == "POST") || (context.Request.HttpMethod == "PUT"))
            {
                CopyStream(context.Request.InputStream, state.RemoteRequest.GetRequestStream());
            }

            // execute remote request
            return state.RemoteRequest.BeginGetResponse(cb, state);
        }

        /// <summary>
        /// Provides an asynchronous process End method when the process ends.
        /// </summary>
        /// <param name="result">Asynchronous result object that contains the status of the process.</param>

        public void EndProcessRequest (IAsyncResult result)
        {
            AsyncState state = result.AsyncState as AsyncState;
            HttpWebResponse response = null;

            try
            {
                response = state.RemoteRequest.EndGetResponse(result) as HttpWebResponse;
            }
            catch (WebException e)
            {
                Log.Error("Request Failed:\n" + state.LocalContext.Request.Url + "\n\n" + e.Message);

                state.LocalContext.Response.StatusCode = 404;

                state.LocalContext.Response.Write("Remote server request failed: " + e.Message);
                state.LocalContext.Response.End();

                return;
            }

            // prepare local response
            state.LocalContext.Response.Buffer = false;
            state.LocalContext.Response.ContentType = response.ContentType;
            state.LocalContext.Response.StatusCode = (int)response.StatusCode;
            state.LocalContext.Response.StatusDescription = response.StatusDescription;

            // forward headers
            foreach (string h in response.Headers)
            {
                if (h == "Location")
                {
                    state.LocalContext.Response.AddHeader(h, response.Headers.Get(h).Replace(state.RemoteWebsite, state.LocalWebsite));
                }
                else if (!WebHeaderCollection.IsRestricted(h) && (h != "Content-Encoding"))
                {
                    state.LocalContext.Response.AddHeader(h, response.Headers.Get(h));
                }
            }

            // forward content
            if (MatchContentType(response.ContentType, ConfigurationManager.AppSettings.Get("ReverseContentTypes")))
            {
                Log.Information("End Request (reversing URLs) ---\nLocal Request: " + state.LocalContext.Request.Url + "\n\nStatus Code: " + response.StatusCode + "\nContent Type: " + response.ContentType + "\nContent Length: " + response.ContentLength);

                using (StreamReader r = new StreamReader(response.GetResponseStream()))
                {
                    string s = r.ReadToEnd();

                    s = s.Replace(state.RemoteWebsite, state.LocalWebsite);

                    if (state.LocalContext.Request.ApplicationPath != "/")
                    {
                        string[] prefixes = ConfigurationManager.AppSettings.Get("ReverseBasePrefixes").Split(',');

                        foreach (string p in prefixes)
                        {
                            if (!string.IsNullOrEmpty(p))
                            {
                                s = s.Replace(p + "\"/", p + "\"" + state.LocalContext.Request.ApplicationPath + "/");
                                s = s.Replace(p + "'/", p + "'" + state.LocalContext.Request.ApplicationPath + "/");
                            }
                        }
                    }

                    state.LocalContext.Response.Write(s);
                }
            }
            else
            {
                Log.Information("End Request (pass-through) ---\nLocal Request: " + state.LocalContext.Request.Url + "\n\nStatus Code: " + response.StatusCode + "\nContent Type: " + response.ContentType + "\nContent Length: " + response.ContentLength);

                using (Stream s = response.GetResponseStream())
                {
                    CopyStream(s, state.LocalContext.Response.OutputStream);
                }
            }

            response.Close();

            state.LocalContext.Response.End();
        }

        /// <summary>
        /// Returns false to prevent this instance from being reused by other requests.
        /// </summary>

        public bool IsReusable
        {
            get
            {
                return true;
            }
        }

        /// <summary>
        /// Inherited from IHttpHandler.
        /// </summary>
        /// <param name="context"></param>

        public void ProcessRequest (HttpContext context)
        {
            // never called
        }

        #endregion


        #region Static

        /// <summary>
        /// Copies the contents of one stream to another.
        /// </summary>
        /// <param name="input">Stream to copy from.</param>
        /// <param name="output">Stream to copy to.</param>

        static public void CopyStream (Stream input, Stream output)
        {
            if ((input != null) && (output != null))
            {
                Byte[] buffer = new byte[1024];
                int bytes = 0;

                while ((bytes = input.Read(buffer, 0, 1024)) > 0)
                {
                    output.Write(buffer, 0, bytes);
                }
            }
        }

        /// <summary>
        /// Checks whether a given path matches one of the specified filter extensions.
        /// </summary>
        /// <param name="path">Content type to check.</param>
        /// <param name="filter">Comma separated list of content types to filter.</param>
        /// <returns>True if the content type is in the list, false otherwise.</returns>

        static public bool MatchContentType (string type, string filter)
        {
            string[] a = filter.Split(',');

            foreach (string s in a)
            {
                if (!string.IsNullOrEmpty(s) && type.StartsWith(s))
                {
                    return true;
                }
            }

            return false;
        }

        #endregion
    }
}